import dataclasses
import logging
from copy import deepcopy
from typing import Any, Mapping, Optional, Sequence

from mutalyzer.description import Description

from .bed import Bed
from .exonviz import draw
from .mutalyzer import (
    Therapy,
    Variant,
    generate_therapies,
    init_description,
    mutation_to_cds_effect,
    protein_prediction,
    sequence_from_description,
)

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Comparison:
    name: str
    percentage: float
    basepairs: str

    @classmethod
    def from_dict(cls, dict: Mapping[str, Any]) -> "Comparison":
        """Create a Variant object from a dict representation of a Variant"""
        return cls(**dict)


@dataclasses.dataclass
class Result:
    """To hold results for separate mutations of a transcript"""

    therapy: Therapy
    comparison: Sequence[Comparison]

    def __gt__(self, other: "Result") -> bool:
        """Sort Result based on the sum of the percentage"""
        if not isinstance(other, Result):
            msg = f"Unsupported comparison between Bed and {type(other)}"
            raise NotImplementedError(msg)

        total_self = sum(c.percentage for c in self.comparison)
        total_other = sum(c.percentage for c in other.comparison)
        return total_self > total_other

    @classmethod
    def from_dict(cls, dict: Mapping[str, Any]) -> "Result":
        return cls(
            therapy=Therapy.from_dict(dict["therapy"]),
            comparison=[Comparison.from_dict(x) for x in dict["comparison"]],
        )


class Transcript:
    def __init__(self, exons: Bed, coding_exons: Optional[Bed] = None):
        self.exons = exons

        if coding_exons is None:
            self.coding_exons = Bed()
        else:
            self.coding_exons = coding_exons

    def records(self) -> Sequence[Bed]:
        """Return the Bed records that make up the Transcript"""
        return [self.exons, self.coding_exons]

    def rna_records(self) -> Sequence[Bed]:
        """Return the Bed records that contain RNA features"""
        return [self.exons]

    def protein_records(self) -> Sequence[Bed]:
        """Return the Bed records that contain protein features"""
        return [self.coding_exons]

    def intersect(self, selector: Bed) -> None:
        """Update transcript to only contain features that intersect the selector"""
        for record in self.records():
            record.intersect(selector)

    def overlap(self, selector: Bed) -> None:
        """Update transcript to only contain features that overlap the selector"""
        for record in self.records():
            record.overlap(selector)

    def subtract(self, selector: Bed) -> None:
        """Remove all features from transcript that intersect the selector"""
        for record in self.records():
            record.subtract(selector)

    def exon_skip(self, selector: Bed) -> None:
        """Remove the exon(s) that overlap the selector from the transcript"""
        exons_to_skip = deepcopy(self.exons)
        exons_to_skip.overlap(selector)
        self.subtract(exons_to_skip)

    def compare(self, other: object) -> Sequence[Comparison]:
        """Compare the size of each record in the transcripts"""
        if not isinstance(other, Transcript):
            raise NotImplementedError

        # Compare each record that makes up self and other
        # The comparison will fail if the record.name does not match
        cmp = list()
        for record1, record2 in zip(self.records(), other.records()):
            percentage = record1.compare(record2)
            fraction = record1.compare_basepair(record2)
            C = Comparison(record1.name, percentage, fraction)
            cmp.append(C)

        return cmp

    def compare_score(self, other: object) -> float:
        """Compare the size of each records in the transcripts

        Returns the average value for all records
        """
        if not isinstance(other, Transcript):
            raise NotImplementedError
        cmp = self.compare(other)

        values = [x.percentage for x in cmp]
        return sum(values) / len(cmp)

    def mutate(self, d: Description, variants: Sequence[Variant]) -> None:
        """Mutate the transcript based on the specified variants"""
        # Update protein features
        protein_changes = Bed.from_blocks(
            self.coding_exons.chrom, *mutation_to_cds_effect(d, variants)
        )
        for record in self.protein_records():
            record.subtract(protein_changes)

        # Update RNA features
        rna_changes = Bed.from_blocks(
            self.exons.chrom, *[v.genomic_coordinates(d) for v in variants]
        )
        for record in self.rna_records():
            self.subtract(rna_changes)

    def analyze(self, hgvs: str) -> Sequence[Result]:
        """Analyze the transcript based on the specified HGVS description"""

        # r. notations are not supported
        coordinate_system = hgvs.split(":")[1][0:1]
        if coordinate_system != "c":
            raise NotImplementedError(
                f"Coordinate system '{coordinate_system}' is not supported"
            )
        # Initialize the input HGVS description
        d = init_description(hgvs)

        # Extract the input variants as internal delins
        sequence = sequence_from_description(d)
        input_variants = [
            Variant.from_model(delins, sequence=sequence)
            for delins in d.delins_model["variants"]
        ]

        results = list()

        # Store the wildtype
        wt = Therapy(
            name="Wildtype",
            hgvsc=hgvs.split("c.")[0] + "c.=",
            hgvsr=hgvs.split("c.")[0] + "r.=",
            hgvsp=protein_prediction(d, [])[0],
            description="These are the annotations as defined on the reference. They are always 100% by definition.",
            variants=list(),
        )
        wildtype = Result(wt, self.compare(self))
        results.append(wildtype)

        # Store the input variants as Therapy
        input = Therapy(
            name="Input",
            hgvsc=hgvs,
            hgvsr=hgvs.replace(":c.", ":r."),
            hgvsp=protein_prediction(d, input_variants)[0],
            description="The annotations based on the supplied input variants.",
            variants=input_variants,
            figure=draw(d),
        )
        patient = deepcopy(self)
        patient.mutate(d, input.variants)
        results.append(Result(input, patient.compare(self)))

        # Generate all possible therapies
        for therapy in generate_therapies(d):
            logger.debug(f"{therapy.name}")

            # Apply the combination to the wildtype transcript
            modified_transcript = deepcopy(self)
            modified_transcript.mutate(d, therapy.variants)
            results.append(Result(therapy, modified_transcript.compare(self)))

        # Order the results
        wt_patient = results[:2]
        rest = sorted(results[2:], reverse=True)
        return wt_patient + rest



from typing import Any, Mapping, Sequence, Union, cast

from pydantic import BaseModel

from .provider import Provider

Payload = Mapping[str, Any]


class Links(BaseModel):
    omim_ids: Sequence[str]
    gene_symbol: str
    ensembl_gene_id: str
    uniprot: str
    decipher: str
    genomic_variant: str
    hgnc: str
    ucsc: str

    databases: Sequence[str] = [
        "omim",
        "lovd",
        "gtex",
        "uniprot",
        "decipher",
        "clinvar",
        "hgnc",
        "ucsc",
        "gnomad",
        "stringdb",
    ]

    def url(self, field: str) -> Union[str, Sequence[str]]:
        if field == "omim":
            urls = list()
            for id in self.omim_ids:
                url = f"https://www.omim.org/entry/{id}"
                urls.append(url)
            return urls
        elif field == "lovd":
            return f"https://databases.lovd.nl/shared/genes/{self.gene_symbol}"
        elif field == "gtex":
            return f"https://gtexportal.org/home/gene/{self.ensembl_gene_id}"
        elif field == "uniprot":
            return f"https://www.uniprot.org/uniprotkb/{self.uniprot}/entry"
        elif field == "decipher":
            return f"https://www.deciphergenomics.org/sequence-variant/{self.decipher}"
        elif field == "clinvar":
            return f"https://www.ncbi.nlm.nih.gov/clinvar/?term={self.genomic_variant}"
        elif field == "hgnc":
            return f"https://www.genenames.org/data/gene-symbol-report/#!/hgnc_id/{self.hgnc}"
        elif field == "ucsc":
            return f"https://genome.cse.ucsc.edu/cgi-bin/hgGene?hgg_gene={self.ucsc}"
        elif field == "gnomad":
            return f"https://gnomad.broadinstitute.org/variant/{self.decipher}?dataset=gnomad_r4"
        elif field == "stringdb":
            return f"https://string-db.org/cgi/network?identifiers={self.gene_symbol}"
        else:
            raise NotImplementedError(f"Unknown field: '{field}'")

    def url_dict(self) -> Mapping[str, str]:
        """Create a flat dict with urls to all databases"""
        d = dict()

        for field in self.databases:
            # omim can contain a list of IDs
            if field == "omim":
                for i, url in enumerate(self.url(field), 1):
                    d[f"{field}_{i}"] = url
            else:
                d[field] = cast(str, self.url(field))

        return d


def guess_refseq_ensembl(variant: str) -> str:
    """Guess if a transcript is RefSeq or Ensembl"""
    if variant.startswith("ENS"):
        return "ensembl"
    else:
        return "refseq"


def lookup_variant(provider: Provider, variant: str, assembly: str = "hg38") -> Links:
    if guess_refseq_ensembl(variant) == "ensembl":
        url = f"https://rest.variantvalidator.org/VariantValidator/variantvalidator_ensembl/{assembly}/{variant}/mane_select?content-type=application/json"
    else:
        url = f"https://rest.variantvalidator.org/VariantValidator/variantvalidator/{assembly}/{variant}/mane_select?content-type=application/json"

    payload = provider.get(url)

    d = parse_payload(payload, variant, assembly)
    d["uniprot"] = lookup_uniprot(provider, d["ensembl_gene_id"])

    return Links(**d)


def lookup_uniprot(provider: Provider, ensembl_gene_id: str) -> str:
    url = f"https://mygene.info/v3/gene/{ensembl_gene_id}?fields=uniprot"
    uniprot_id: str = provider.get(url)["uniprot"]["Swiss-Prot"]
    return uniprot_id


def extract_variant(payload: Payload, variant: str) -> Payload:
    """Extract the variant section from the payload"""
    for value in payload.values():
        # Skip flag field
        if not isinstance(value, dict):
            continue
        submitted_variant = value.get("submitted_variant")
        if submitted_variant == variant:
            var_payload: Payload = value
            return var_payload
    else:
        msg = f"Unable to parse VariantValidator output, '{variant}' not found."
        raise ValueError(msg)


def parse_payload(payload: Payload, variant: str, assembly: str) -> dict[str, Any]:
    # Check the flag to see if the reply is valid
    flag = payload["flag"]
    if flag == "warning":
        w = payload.get("validation_warning_1", dict())
        errors = "\n".join(w.get("validation_warnings", []))
        raise ValueError(errors)
    if flag != "gene_variant":
        msg = f"Unknown VariantValidator flag: {flag}"
        raise NotImplementedError(msg)

    var = extract_variant(payload, variant)
    d = {
        "omim_ids": var["gene_ids"]["omim_id"],
        "gene_symbol": var["gene_symbol"],
        "ensembl_gene_id": var["gene_ids"]["ensembl_gene_id"],
        "hgnc": var["annotations"]["db_xref"]["hgnc"],
        "ucsc": var["gene_ids"]["ucsc_id"],
        "genomic_variant": var["primary_assembly_loci"][assembly][
            "hgvs_genomic_description"
        ],
    }
    # Get the 'VCF' notation on the specified assembly
    vcf = var["primary_assembly_loci"][assembly]["vcf"]
    # Remove the 'chr' prefix
    vcf["chr"] = vcf["chr"][3:]
    # Decypher uses {chrom}-{pos}-{ref}-{alt} as variant ID
    d["decipher"] = "-".join(vcf[field] for field in ["chr", "pos", "ref", "alt"])
    return d



import logging
from typing import Any, Mapping

from .models import Assembly, EnsemblTranscript
from .provider import Provider

logger = logging.getLogger(__name__)

ENSEMBL_TO_UCSC = {
    Assembly.HUMAN: "hg38",
    Assembly.RAT: "rn6",
}


def chrom_to_uscs(seq_region_name: str) -> str:
    return "chrM" if seq_region_name == "MT" else f"chr{seq_region_name}"


def ucsc_url(transcript: EnsemblTranscript, track: str = "knownGene") -> str:
    genome = ENSEMBL_TO_UCSC[transcript.assembly_name]
    url = ";".join(
        (
            f"https://api.genome.ucsc.edu/getData/track?genome={genome}",
            f"chrom={chrom_to_uscs(transcript.seq_region_name)}",
            f"track={track}",
            f"start={transcript.start}",
            f"end={transcript.end}",
        )
    )

    return url


def lookup_knownGene(
    provider: Provider, transcript: EnsemblTranscript, track_name: str
) -> Mapping[str, Any]:
    url = ucsc_url(transcript, track_name)
    track = provider.get(url)
    ts = f"{transcript.id}.{transcript.version}"
    track[track_name] = [
        entry for entry in track[track_name] if entry.get("name") == ts
    ]
    return track



from typing import Any, Iterator, Optional, Sequence

from .range import Range, intersect, overlap, subtract

# colorRgb field from Bed
color = tuple[int, ...]


class Bed:
    def __init__(
        self,
        chrom: Optional[str] = None,
        chromStart: Optional[int] = None,
        chromEnd: Optional[int] = None,
        name: str = ".",
        score: int = 0,
        strand: str = ".",
        thickStart: Optional[int] = None,
        thickEnd: Optional[int] = None,
        itemRgb: color = (0, 0, 0),
        blockCount: Optional[int] = None,
        blockSizes: Optional[list[int]] = None,
        blockStarts: Optional[list[int]] = None,
        **ignored: Any,
    ) -> None:
        # Required attributes for .bed files
        self.chrom = chrom if chrom is not None else ""
        self.chromStart = chromStart if chromStart is not None else 0
        self.chromEnd = chromEnd if chromEnd is not None else 0

        # Simple attributes
        self.name = name
        self.score = score
        self.strand = strand

        if thickStart is None:
            self.thickStart = self.chromStart
        else:
            self.thickStart = thickStart

        if thickEnd is None:
            self.thickEnd = self.chromEnd
        else:
            self.thickEnd = thickEnd

        self.itemRgb = itemRgb

        # Set the blocks
        if blockSizes is None:
            self.blockSizes = [self.chromEnd - self.chromStart]
        else:
            self.blockSizes = blockSizes

        if blockStarts is None:
            # blockStarts are relative to chromStart, and the first block must
            # start at 0
            self.blockStarts = [0]
        else:
            self.blockStarts = blockStarts

        self.blockCount = blockCount if blockCount else len(self.blockSizes)

        self.validate()

    def validate(self) -> None:
        """Validate the internal constistence of the Bed record"""
        if self.thickStart < self.chromStart or self.thickStart > self.chromEnd:
            raise ValueError(f"thickStart outside of record ({self})")
        if self.thickEnd < self.chromStart or self.thickEnd > self.chromEnd:
            raise ValueError(f"thickEnd outside of record ({self})")
        if self.thickEnd < self.thickStart:
            raise ValueError("thickEnd before thickStart")
        if len(self.blockSizes) != self.blockCount:
            msg = f"blockCount({self.blockCount=}) does not match the number of blocks({self.blockSizes=})"
            raise ValueError(msg)
        if len(self.blockSizes) != len(self.blockStarts):
            msg = f"number of values differs between blockSizes({self.blockSizes}) and blockStarts({self.blockStarts})"
            raise ValueError(msg)

        # Initialise with the end of the first block
        prev_end = self.chromStart + self.blockStarts[0] + self.blockSizes[0]
        prev_start = self.blockStarts[0]
        blocks = list(self.blocks())[1:]
        for start, end in blocks:
            if start < prev_start:
                raise ValueError("Blocks must be in ascending order")
            elif start < prev_end:
                raise ValueError("Blocks must not overlap")
            else:
                prev_end = end
                prev_start = start

        # The first block must start at chromStart
        if self.blockStarts[0] != 0:
            raise ValueError("The first block must start at chromStart")

        # The last block must end at chromEnd
        block_end = self.blockStarts[-1] + self.blockSizes[-1] + self.chromStart
        if block_end != self.chromEnd:
            raise ValueError(f"Last block({block_end=}) must end at {self.chromEnd=}")

    def blocks(self) -> Iterator[tuple[int, int]]:
        """Iterate over all blocks in the Bed record"""
        for size, start in zip(self.blockSizes, self.blockStarts):
            block_start = self.chromStart + start
            block_end = block_start + size
            yield (block_start, block_end)

    def __str__(self) -> str:
        return "\t".join(
            map(
                str,
                (
                    self.chrom,
                    self.chromStart,
                    self.chromEnd,
                    self.name,
                    self.score,
                    self.strand,
                    self.thickStart,
                    self.thickEnd,
                    ",".join(map(str, self.itemRgb)),
                    self.blockCount,
                    ",".join(map(str, self.blockSizes)),
                    ",".join(map(str, self.blockStarts)),
                ),
            )
        )

    def __repr__(self) -> str:
        return (
            f"Bed({self.chrom}, "
            f"{self.chromStart}, {self.chromEnd}, "
            f"name='{self.name}', "
            f"score={self.score}, "
            f"strand='{self.strand}', "
            f"thickStart='{self.thickStart}', "
            f"thickEnd='{self.thickEnd}', "
            f"blockCount='{self.blockCount}', "
            f"blockSizes='{self.blockSizes}', "
            f"blockStarts='{self.blockStarts}', "
            ")"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Bed):
            msg = f"Unsupported comparison between Bed and {type(other)}"
            raise NotImplementedError(msg)
        return all(
            (
                self.chrom == other.chrom,
                self.chromStart == other.chromStart,
                self.chromEnd == other.chromEnd,
                self.name == other.name,
                self.score == other.score,
                self.strand == other.strand,
                self.thickStart == other.thickStart,
                self.thickEnd == other.thickEnd,
                self.itemRgb == other.itemRgb,
                self.blockCount == other.blockCount,
                self.blockSizes == other.blockSizes,
                self.blockStarts == other.blockStarts,
            )
        )

    def __bool__(self) -> bool:
        return self.size > 0

    def _zero_out(self) -> None:
        """Zero out the Bed object, by setting all ranges to the start"""
        self.chromEnd = self.thickStart = self.thickEnd = self.chromStart

        self.blockCount = 1
        self.blockSizes = self.blockStarts = [0]

    @staticmethod
    def _csv_to_int(csv: str) -> list[int]:
        """Convert a csv list to a list of integers"""
        integers = list()
        for num in csv.split(","):
            if num != "":
                integers.append(int(num))
        return integers

    @classmethod
    def from_bedfile(cls, line: str) -> "Bed":
        """Create a BED record from a line from a Bed file"""
        header = [
            "chrom",
            "chromStart",
            "chromEnd",
            "name",
            "score",
            "strand",
            "thickStart",
            "thickEnd",
            "itemRgb",
            "blockCount",
            "blockSizes",
            "blockStarts",
        ]

        # Parse the bed fields into a dict
        d = {k: v for k, v in zip(header, line.split("\t"))}

        return cls(
            chrom=d["chrom"],
            chromStart=int(d["chromStart"]),
            chromEnd=int(d["chromEnd"]),
            name=d.get("name", ""),
            score=int(d.get("score", 0)),
            strand=d.get("strand", "."),
            thickStart=int(d["thickStart"]) if "thickStart" in d else None,
            thickEnd=int(d["thickEnd"]) if "thickEnd" in d else None,
            itemRgb=(
                tuple(cls._csv_to_int(d["itemRgb"])) if "itemRgb" in d else (0, 0, 0)
            ),
            blockCount=int(d["blockCount"]) if "blockCount" in d else None,
            blockSizes=cls._csv_to_int(d["blockSizes"]) if "blockSizes" in d else None,
            blockStarts=(
                cls._csv_to_int(d["blockStarts"]) if "blockStarts" in d else None
            ),
        )

    @classmethod
    def from_blocks(cls, chrom: str, *blocks: Range) -> "Bed":
        """Create a Bed record from multiple Ranges"""
        bed = cls(chrom, 0, 0)

        # Ensure the blocks are sorted in ascending order
        sorted_blocks = sorted(blocks, key=lambda x: x[0])
        bed.update(sorted_blocks)

        # Set the thickStart, thickEnd to the entire range
        bed.thickStart = bed.chromStart
        bed.thickEnd = bed.chromEnd

        return bed

    def intersect(self, other: object) -> None:
        """Update record to only contain features that overlap other"""
        if not isinstance(other, Bed):
            raise NotImplementedError

        if self.strand != other.strand:
            raise ValueError("Conflicting strands, intersection not possible")

        # If other is on a different chromosome, we zero out self since there
        # is no overlap
        if self.chrom != other.chrom:
            self._zero_out()
            return

        # Determine all intersected ranges
        intersected: list[Range] = list()

        for range1 in self.blocks():
            for intersector in other.blocks():
                intersected += intersect(range1, intersector)

        self.update(intersected)

    def overlap(self, other: object) -> None:
        """All blocks from self that (partially) overlap blocks from other"""
        if not isinstance(other, Bed):
            raise NotImplementedError

        # If other is on a different chromosome, there can be no overlap
        if self.chrom != other.chrom:
            self._zero_out()

        # Calculating overlap on different strands is not supported
        if self.strand != other.strand:
            raise ValueError(
                "Calculating overlap on different strands is not supported"
            )

        blocks_to_keep = list()
        for block in self.blocks():
            for other_block in other.blocks():
                if overlap(block, other_block):
                    blocks_to_keep.append(block)
                    break  # Go to the next block once we find overlap
        self.update(blocks_to_keep)

    def subtract(self, other: object) -> None:
        """Subtract the blocks in other from blocks in self"""
        if not isinstance(other, Bed):
            raise NotImplementedError

        if self.chrom != other.chrom:
            return

        subtracted_blocks = subtract(list(self.blocks()), list(other.blocks()))
        self.update(subtracted_blocks)

    def compare(self, other: object) -> float:
        """
        Compare the size of Bed objects

        Raises a ValueError if chrom, name or strand are not identical
        """

        def get_properties(bed: Bed) -> tuple[str, ...]:
            """Return a tuple of the properties that must match"""
            return (bed.chrom, bed.name, bed.strand)

        if not isinstance(other, Bed):
            raise NotImplementedError

        # Other can not have size zero
        if other.size == 0:
            msg = f"Comparison not allowed with zero-size Bed record ({other=})"
            raise ValueError(msg)

        # We only compare Bed objects if their properties match
        prop_self = get_properties(self)
        prop_other = get_properties(other)

        if prop_self != prop_other:
            msg = "Comparison failed, properties mismatch: {prop_self} != {prop_other}"
            raise ValueError(msg)

        return self.size / other.size

    def compare_basepair(self, other: object) -> str:
        """
        Compare the size of Bed objects, return a string showing remainting basepairs

        Raises a ValueError if chrom, name or strand are not identical
        """

        def get_properties(bed: Bed) -> tuple[str, ...]:
            """Return a tuple of the properties that must match"""
            return (bed.chrom, bed.name, bed.strand)

        if not isinstance(other, Bed):
            raise NotImplementedError

        # Other can not have size zero
        if other.size == 0:
            msg = f"Comparison not allowed with zero-size Bed record ({other=})"
            raise ValueError(msg)

        # We only compare Bed objects if their properties match
        prop_self = get_properties(self)
        prop_other = get_properties(other)

        if prop_self != prop_other:
            msg = "Comparison failed, properties mismatch: {prop_self} != {prop_other}"
            raise ValueError(msg)

        return f"{self.size}/{other.size}"

    def update(self, ranges: Sequence[Range]) -> None:
        """Update a Bed object with a list of ranges"""
        # Check that thickStart/thickEnd have not been set
        if self.thickStart != self.chromStart or self.thickEnd != self.chromEnd:
            raise NotImplementedError

        if not ranges:
            self._zero_out()
            return

        # The ranges are sorted
        self.chromStart = ranges[0][0]
        self.chromEnd = ranges[-1][-1]

        # Set to the new start/end of the record
        self.thickStart = self.chromStart
        self.thickEnd = self.chromEnd

        # Set the number of blocks
        self.blockCount = len(ranges)

        # Set the block starts and sizes
        self.blockSizes = list()
        self.blockStarts = list()

        for r in ranges:
            size, start = _range_to_size_start(range=r, offset=self.chromStart)
            self.blockSizes.append(size)
            self.blockStarts.append(start)
        self.validate()

    @property
    def size(self) -> int:
        """The size of a Bed record is the sum of its blocks"""
        return sum(self.blockSizes)


def _range_to_size_start(range: Range, offset: int) -> tuple[int, int]:
    """Convert a range to size, start

    BED format uses blockSizes and blockStarts to represent ranges
    """

    size = range[1] - range[0]
    start = range[0] - offset
    return size, start



import logging
from typing import Any, Mapping

from .models import EnsemblTranscript
from .provider import Provider

logger = logging.getLogger(__name__)


def lookup_transcript(provider: Provider, transcript_id: str) -> EnsemblTranscript:
    transcript, version = transcript_id.split(".")
    url = (
        f"http://rest.ensembl.org/lookup/id/{transcript}?content-type=application/json"
    )
    ts = provider.get(url)

    _check_transcript(ts, int(version))

    return payload_to_ensemble_transcript(ts)


def payload_to_ensemble_transcript(payload: Mapping[str, Any]) -> EnsemblTranscript:
    return EnsemblTranscript(**payload)


def _check_transcript(payload: Mapping[str, Any], version: int) -> None:
    if (e_version := payload.get("version")) != version:
        msg = f"Ensembl returned a different version of the transcript: {e_version} instead of {version}"
        raise ValueError(msg)
    return



from typing import Sequence

Range = tuple[int, int]


def empty(a: Range) -> bool:
    """Determine if range a is empty"""
    return a[0] == a[1]


def overlap(a: Range, b: Range) -> bool:
    """Determine if ranges a and b overlap"""
    # A and B overlap wether the intersection is not empty
    if intersect(a, b):
        return True
    return False


def intersect(a: Range, b: Range) -> Sequence[Range]:
    """Determine the intersection between ranges a and b"""
    start = max(a[0], b[0])
    end = min(a[1], b[1])

    if start + 1 > end:
        return list()
    else:
        return [(start, end)]


def subtract(a: Sequence[Range], b: Sequence[Range]) -> Sequence[Range]:
    """
    Subtract the regions in b from a
    """

    # Iterate over a and b, and add None as a stop value
    a_iter = iter([x for x in a] + [None])
    b_iter = iter([x for x in b] + [None])

    A = next(a_iter)
    B = next(b_iter)

    results: list[tuple[int, int]] = list()

    while True:
        # If we exhaust A
        if A is None:
            break
        # If A is empty, we skip it
        if empty(A):
            A = next(a_iter)
            continue

        # If we exhaust B
        if B is None:
            # Get the values for A still waiting to be processed
            # Remove the None value
            left_over = [x for x in a_iter if x is not None]
            results += [A] + left_over
            break
        # if B is empty, we skip it
        if empty(B):
            B = next(b_iter)
            continue

        # A before B
        if A[1] <= B[0]:
            results.append(A)
            A = next(a_iter)
            continue

        # A after B
        if A[0] >= B[1]:
            B = next(b_iter)
            continue

        # If we are here, A and B overlap
        o = intersect(A, B)[0]

        # B is in the middle of A
        if B == o:
            results.append((A[0], o[0]))
            A = (o[1], A[1])
        # B fully overlaps A
        elif A == o:
            A = next(a_iter)
        # B overlaps start of A
        elif o[1] == B[1]:
            B = next(b_iter)
            A = (o[1], A[1])
        # B overlaps the end of A
        elif A[1] == o[1]:
            results.append((A[0], o[0]))
            A = next(a_iter)

    # Remove ranges of size zero
    results = [x for x in results if x[1] > x[0]]
    return results



from typing import Mapping, Sequence

import uvicorn as uvicorn
from fastapi import Body, FastAPI, HTTPException
from fastapi.responses import RedirectResponse
from typing_extensions import Annotated

from .models import BedModel, TranscriptId, TranscriptModel
from .mutalyzer import HGVS
from .provider import Provider
from .transcript import Comparison, Result
from .variant_validator import lookup_variant
from .wrappers import lookup_transcript

app = FastAPI()
provider = Provider()


@app.get("/")
async def redirect() -> RedirectResponse:
    response = RedirectResponse(url="/docs")
    return response


@app.post("/links")
async def get_links(variant: HGVS) -> Mapping[str, str]:
    """Lookup external references for the specified variant"""
    try:
        reply = lookup_variant(provider, variant.description).url_dict()
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))
    return reply


@app.post("/transcript")
async def get_transcript(transcript: TranscriptId) -> TranscriptModel:
    """Lookup the specified transcript"""
    try:
        ts = lookup_transcript(provider, transcript.id)
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))
    return ts


@app.post("/transcript/exonskip")
async def exon_skip(transcript: TranscriptModel, region: BedModel) -> TranscriptModel:
    """Skip exons that overlap the specified region"""
    ts = transcript.to_transcript()
    skip_region = region.to_bed()
    ts.exon_skip(skip_region)
    return TranscriptModel.from_transcript(ts)


@app.post("/transcript/compare")
async def compare(
    self: Annotated[
        TranscriptModel,
        Body(
            examples=[
                {
                    "exons": {
                        "chrom": "chr1",
                        "blocks": [[0, 10], [50, 60], [70, 100]],
                        "name": "exons",
                        "score": 0,
                        "strand": ".",
                    },
                    "cds": {
                        "chrom": "chr1",
                        "blocks": [[40, 72]],
                        "name": "cds",
                        "score": 0,
                        "strand": ".",
                    },
                }
            ]
        ),
    ],
    other: TranscriptModel,
) -> Sequence[Comparison]:
    """Compare two transcripts"""
    s = self.to_transcript()
    o = other.to_transcript()

    return s.compare(o)


@app.post("/hgvs/analyze", response_model=None)
async def analyze(hgvs: HGVS) -> Sequence[Result]:
    """Analyze all possible exons skips for the spcified HGVS variant"""
    transcript_id = hgvs.description.split(":")[0]
    transcript_model = lookup_transcript(provider, transcript_id)
    transcript = transcript_model.to_transcript()
    return transcript.analyze(hgvs.description)



from .ensembl import lookup_transcript as lookup_transcript_ens
from .models import BedModel, TranscriptModel
from .provider import Provider
from .ucsc import lookup_knownGene


def lookup_transcript(provider: Provider, transcript_id: str) -> TranscriptModel:
    r = lookup_transcript_ens(provider, transcript_id)
    # track_name = "ncbiRefSeq"
    # track_name = "ensGene"
    track_name = "knownGene"
    track = lookup_knownGene(provider, r, track_name)
    knownGene = track[track_name][0]
    exons = BedModel.from_ucsc(knownGene)

    # Rename the exons track to "Exons"
    exons.name = "Exons"

    # The CDS is defied as the thickStart, thickEnd in ucsc
    chrom = knownGene["chrom"]
    start = knownGene["thickStart"]
    end = knownGene["thickEnd"]
    name = "CDS"
    strand = knownGene["strand"]
    cds = BedModel(
        chrom=chrom, blocks=[(start, end)], name=name, strand=strand
    ).to_bed()

    # Determine the coding region
    coding_exons = exons.to_bed()
    coding_exons.name = "Coding exons"
    coding_exons.intersect(cds)

    return TranscriptModel(exons=exons, coding_exons=BedModel.from_bed(coding_exons))



"""
Module that contains the command line app, so we can still import __main__
without executing side effects
"""

import argparse
import dataclasses
import json
import logging
import os
import secrets

from mutalyzer.description import Description

from gtgt.flask import render
from gtgt.transcript import Result

from .mutalyzer import (
    Variant,
    generate_therapies,
    init_description,
    sequence_from_description,
)
from .provider import Provider
from .variant_validator import lookup_variant
from .wrappers import lookup_transcript


def set_logging(level: str) -> None:
    format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    format = "%(name)s - %(levelname)s - %(message)s"
    logging.basicConfig(
        format=format,
        level=level.upper(),
    )


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Description of command.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--log",
        choices=["debug", "info", "warning", "error", "critical"],
        default="info",
    )

    subparsers = parser.add_subparsers(dest="command")
    parser.add_argument("--cachedir", type=str, default=os.environ.get("GTGT_CACHE"))

    transcript_parser = subparsers.add_parser(
        "transcript", help="Transcript Information"
    )

    transcript_parser.add_argument(
        "transcript_id", type=str, help="Transcript of interest"
    )

    link_parser = subparsers.add_parser("links", help="Links to external resources")
    link_parser.add_argument("hgvs_variant", type=str, help="Variant of interest")

    api_server_parser = subparsers.add_parser(
        "api_server", help="Run the GTGT API server"
    )
    api_server_parser.add_argument(
        "--host", default="localhost", help="Hostname to listen on"
    )

    web_server_parser = subparsers.add_parser(
        "webserver", help="Run the GTGT web server"
    )
    web_server_parser.add_argument(
        "--host", default="localhost", help="Hostname to listen on"
    )
    web_server_parser.add_argument(
        "--debug", default=False, action="store_true", help="Run Flask in debug mode"
    )

    mutator_parser = subparsers.add_parser(
        "mutate", help="Mutate the specified transcript"
    )

    mutator_parser.add_argument("transcript_id", help="The transcript to mutate")

    analyze_parser = subparsers.add_parser(
        "analyze", help="Analyze all possible exons skips for the spcified HGVS variant"
    )
    analyze_parser.add_argument(
        "hgvs", help="HGVS description of the transcript of interest"
    )

    export_parser = subparsers.add_parser(
        "export", help="Export the specified Transcript to BED format"
    )
    export_parser.add_argument(
        "hgvs", help="HGVS description of the transcript of interest"
    )

    render_parser = subparsers.add_parser(
        "render", help="Render the HTML template (helper)"
    )

    render_parser.add_argument("--results", type=str, default="")
    render_parser.add_argument("--error", "-e", help="Error payload")
    render_parser.add_argument("--variant", type=str, default="10A>T")
    args = parser.parse_args()

    set_logging(args.log)
    logger = logging.getLogger(__name__)

    provider = Provider(args.cachedir)

    if args.command == "transcript":
        ts = lookup_transcript(provider, args.transcript_id)
        print(ts.model_dump_json())
    elif args.command == "links":
        logger.debug(args)
        links = lookup_variant(provider, args.hgvs_variant).url_dict()
        for website, url in links.items():
            print(f"{website}: {url}")
    elif args.command == "api_server":
        try:
            from .app import app, uvicorn
        except ModuleNotFoundError:
            logger.critical(
                "Missing modules, please install with 'pip install gtgt[api_server]'"
            )
            exit(-1)
        uvicorn.run(app, host=args.host)
    elif args.command == "mutate":
        desc = f"{args.transcript_id}:c.="
        d = init_description(desc)
        for therapy in generate_therapies(d):
            print(f"{therapy.name}: {therapy.hgvsc}")
    elif args.command == "analyze":
        transcript_id = args.hgvs.split(":")[0]
        transcript_model = lookup_transcript(provider, transcript_id)
        transcript = transcript_model.to_transcript()
        # Convert Result objects to dict
        results = [dataclasses.asdict(x) for x in transcript.analyze(args.hgvs)]
        print(json.dumps(results, indent=True, default=vars))
    elif args.command == "webserver":
        try:
            from .flask import app as flask_app
        except ModuleNotFoundError:
            logger.critical(
                f"Missing modules, please install with 'pip install gtgt[webserver]'"
            )
            exit(-1)
        if not flask_app.config.get("SECRET_KEY"):
            flask_app.secret_key = secrets.token_hex()
        flask_app.run(args.host, debug=args.debug)
    elif args.command == "export":
        # Get the transcript
        transcript_id = args.hgvs.split(":")[0]
        transcript_model = lookup_transcript(provider, transcript_id)
        transcript = transcript_model.to_transcript()

        # Mutate the transcript
        d = init_description(args.hgvs)
        sequence = sequence_from_description(d)
        input_variants = [
            Variant.from_model(delins, sequence=sequence)
            for delins in d.delins_model["variants"]
        ]
        transcript.mutate(d, input_variants)

        for record in transcript.records():
            print(record)
    elif args.command == "render":
        if args.results:
            with open(args.results) as fin:
                file_payload = [Result.from_dict(x) for x in json.load(fin)]
        else:
            file_payload = []

        template_file = "templates/index.html.j2"
        print(
            render(
                template_file,
                variant=args.variant,
                results=file_payload,
                error=args.error,
            )
        )
    else:
        raise NotImplementedError()


if __name__ == "__main__":
    main()
